package application;

// --------- JOINPOINTTYPES

joinpointtype CheckingOut {
	// going to take this item and amount from stock
	Item item;
	int amount;
}


joinpointtype Buying extends CheckingOut {
	// buying this amount of item 
}

joinpointtype Renting extends CheckingOut {
	  // renting this item and amount until returndate
	  java.util.Date returndate;
}

// --------- CLASSES

class ShoppingSession exhibits CheckingOut,Buying {
	pointcut CheckingOut : call(* add(..)) && args(item, amount,..);
	pointcut Buying : call(* add(..)) && args(item, amount) && withincode(void buy(..));

	ShoppingCart sc = new ShoppingCart();
	Invoice inv = new Invoice();
	Log log = new Log("Shopping-Session-Log");
	Customer cus = customerLogOn();
	int totalAmount = 0;
	void buy(Item item, int amount) {
		System.out.println("ShoppingSession.buy("+item+","+amount+")");
	    sc.add(item, amount); // Matched by CheckingOut, Buying
	    inv.add(item, amount, cus); 
	    log.add(item, amount, cus); 
	    exhibit new Buying(item, amount) { // Matched by Buying
	      totalAmount += amount;
	    };
	    System.out.println("totalAmount="+totalAmount);
	}
	void rent(Item item, int amount, java.util.Date returndate) {
	    sc.add(item, amount); // Matched by CheckingOut
	    log.rent(item, amount, cus);
	}
	Customer customerLogOn() {return new Customer("Test "+System.currentTimeMillis());}
	
}

class ShoppingCart /*exhibits Buying*/ {
//	  pointcut Buying : execution(* add(Item, int)) && args(item, amount);
	void add(Item it, int amount) {
		System.out.println("ShoppingCart.add("+it+","+amount+")");
	}
}

class Invoice {
	void add(Item it, int amount, Customer cus) {
		System.out.println("Invoice.add("+it+","+amount+","+cus+")");
	}
}

class Log /*exhibits Buying*/ {
//	pointcut Buying : execution(* add(Item, int, ..)) && args(item, amount, ..);
	String log;	
	Log(String log) {
	  this.log = log;
	}
	void add(Item item, int amount, Customer c) {
		System.out.println("Log.add("+item+","+amount+","+c+")");
	  log += "\n\t\t\tcustomer: "
		  + c.name
		  + " bought item: "
		  + Item.names[item.category]
		  + " amount: "
		  + amount;
	  System.out.println(log);
	}
	void rent(Item item, int amount, Customer c) {
		  log += "\n\t\t\tcustomer: "
			  + c.name
			  + " rent item: "
			  + Item.names[item.category]
			  + " amount: "
			  + amount;
		  System.out.println(log);
		}
}

class Customer {
	String name;
	Customer(String name) {
	  this.name = name;
	}
	public String toString() {return "Customer:"+name;}
}

class Item {
	static final int BOOK   = 0;
	static final int CAMERA = 1;
	static final int MP3    = 2;
	static final String[] names = {"book", "camera", "mp3"};
	
	int category;
	
	Item(int cat) {
	  this.category = cat;
	}
	public String toString() {return "Item:"+Item.names[category];}	
}

class OutOfStockException extends RuntimeException {
	private Item i;
	public OutOfStockException(Item i) {
		this.i = i;
	}
}

class Stock {
	static java.util.Map items = new java.util.HashMap();
	static int amount(application.Item i) {
		return 5;		
	}
}

// ---------- ASPECTS -----------------------

aspect BusinessRules advises CheckingOut {
	before(CheckingOut co) {
		System.out.println("\t[check]\tCheckingOut : before "+thisJoinPoint+" -> "+co.getClass());
		System.out.println("\t[check]\tcheck if Stock is sufficient : Stock : "+Stock.amount(co.item)+" amount : "+co.amount);
		if (Stock.amount(co.item) < co.amount) {
			System.out.println("\t[check]\t-> Stock is not sufficient for Item : "+co.item);
			throw new OutOfStockException(co.item);
		}
	}
}

aspect BonusProgram advises CheckingOut/*, Buying*/ {
	
	void around (CheckingOut jp) {
		System.out.println("\t** BonusProgram: around  -> "+jp.getClass());
		if (jp.item.category == Item.BOOK) {
			System.out.println("\t[buy]\t-> Item == BOOK : add "+jp.amount / 2+" to amount "+jp.amount);
			jp.amount += jp.amount / 2;
		}
		proceed(jp);
	}
	
	// This advice is no longer required as the around(CheckingOut) is applied on JP of Type Buying 
//	void around (Buying jp) {
//		System.out.println("\t[buy]\tBonusProgram: around "+thisJoinPoint+" -> "+jp.getClass());
//		if (jp.item.category == Item.BOOK) {
//			System.out.println("\t[buy]\t-> Item == BOOK : add "+jp.amount / 2+" to amount "+jp.amount);
//			jp.amount += jp.amount / 2;
//		}
//		proceed(jp);
//	}
	
}

// ---------- TEST

public class Test {
	public static void main(String[] args) {
		
		ShoppingSession ss = new ShoppingSession();
		ss.buy(new Item(Item.MP3), 1);
		System.out.println("\n------------------\n");
		ss.buy(new Item(Item.BOOK), 1);
		System.out.println("\n------------------\n");
		ss.buy(new Item(Item.BOOK), 2);
		System.out.println("\n------------------\n");
		try {
			ss.buy(new Item(Item.BOOK), 6);
		} catch(OutOfStockException e) {
			System.out.println(e.getClass()+" Excpetion");
		}
		System.out.println("\n------------------\n");
		ss.rent(new Item(Item.BOOK), 2, new java.util.Date(System.currentTimeMillis()+ 1000*60*60*24*3));
		System.out.println("\n------------------\n");
		ss.rent(new Item(Item.MP3), 2, new java.util.Date(System.currentTimeMillis()+ 1000*60*60*24*2));
		System.out.println("\n------------------\n");
		try {
			ss.rent(new Item(Item.BOOK), 6,new java.util.Date(System.currentTimeMillis()+ 1000*60*60*24*7));
		} catch(OutOfStockException e) {
			System.out.println(e.getClass()+" Excpetion");
		}
		System.out.println("\n------------------\n");
	}
}